#!/usr/bin/python

import client
import crypto
import json
import os
import profile
import sys

PATH_FILESYSTEM = "filesystem/"
PATH_SHARE = "share/"

dirs = []

def mkinode(obj, efnm = None, sk = None):
    efnm = crypto.gen() if efnm == None else efnm
    sk = crypto.sgen() if sk == None else sk
    c = crypto.senc(sk, json.dumps(obj))
    hash = crypto.hash(c)
    client.put(efnm, c)
    return [efnm, crypto.hash(efnm), sk, hash]

def updatefilesystem(delete = False):
    contents = []
    root = dirs.index(None) - 1
    updateconsistencys(dirs[root], delete)
    for i in range(root, -1, -1):
        [efnm, hefnm, sk, hash] = profile.profile.roots[dirs[root]] if i == root else contents[0][dirs[i]]
        c = client.get(hefnm)
        if crypto.hash(c) != hash:
            raise Exception("ALARM: Integrity check failed on filesystem!")
        if delete:
            client.delete(efnm)
        contents.insert(0, json.loads(crypto.sdec(sk, c)))
    return contents

def maskshare(tuple, condition):
    return [tuple[0] if condition else None, tuple[1], tuple[2], tuple[3]]

def sendshare(username, root, permission):
    with open(PATH_SHARE + username, "a") as ouf:
        if permission == "none":
            ouf.write(json.dumps([root, None, None, None, None]) + os.linesep)
        else:
            ouf.write(json.dumps([root, maskshare(profile.profile.roots[root], permission == "write"), maskshare(profile.profile.permissions[root], username == profile.profile.username), profile.profile.consistencys[root], profile.profile.stored[root]]) + os.linesep)

def receiveshare():
    if os.path.isfile(PATH_SHARE + profile.profile.username):
        with open(PATH_SHARE + profile.profile.username, "r") as inf:
            for line in inf:
                [dir, root, permission, consistency, stored] = json.loads(line[:-1])
                if root == None:
                    del profile.profile.roots[dir]
                    del profile.profile.permissions[dir]
                    del profile.profile.consistencys[dir]
                    del profile.profile.stored[dir]
                else:
                    profile.profile.roots[dir] = root
                    profile.profile.permissions[dir] = permission
                    profile.profile.consistencys[dir] = consistency
                    profile.profile.stored[dir] = stored
        os.remove(PATH_SHARE + profile.profile.username)

def updatepermissions(delete = False):
    root = dirs[dirs.index(None) - 1]
    updateconsistencys(root, delete)
    c = client.get(profile.profile.permissions[root][1])
    if crypto.hash(c) != profile.profile.permissions[root][3]:
        raise Exception("ALARM: Integrity check failed on permission file!")
    if delete == True:
        client.delete(profile.profile.permissions[root][0])
    return json.loads(crypto.sdec(profile.profile.permissions[root][2], c))

def post(contents, permissions = None):
    root = dirs.index(None) - 1
    if contents != None:
        for i in range(0, root):
            contents[i + 1][dirs[i]] = mkinode(contents[i])
        profile.profile.roots[dirs[root]] = mkinode(contents[root], profile.profile.roots[dirs[root]][0], profile.profile.roots[dirs[root]][2])
    if permissions != None:
        profile.profile.permissions[dirs[root]] = mkinode(permissions, profile.profile.permissions[dirs[root]][0], profile.profile.permissions[dirs[root]][2])
    postconsistencys(dirs[root])

def updateconsistencys(root, delete):
    receiveshare()
    history = json.loads(crypto.sdec(profile.profile.consistencys[root][2], client.get(profile.profile.consistencys[root][1])))
    timestamp = None
    for consistency in history:
        if (timestamp != None and consistency[0] != timestamp + 1) or (profile.profile.stored[root][0][0] == consistency[0] and profile.profile.stored[root][0] != consistency):
            raise Exception("ALARM: Integrity check failed on consistency file!")
        elif profile.profile.stored[root][0][0] < consistency[0]:
            profile.profile.stored[root][1] = crypto.hash(json.dumps((consistency, profile.profile.stored[root][1])))
            profile.profile.stored[root][0] = consistency
        timestamp = consistency[0]
    if timestamp == None or timestamp < profile.profile.stored[root][0][0]:
        raise Exception("ALARM: Integrity check failed on consistency file!")
    profile.profile.roots[root][3] = profile.profile.stored[root][0][3]
    profile.profile.permissions[root][3] = profile.profile.stored[root][0][5]
    if delete == False:
        postconsistencys(root)

def postconsistencys(root):
    profile.profile.stored[root][0] = [profile.profile.stored[root][0][0] + 1, profile.profile.username, profile.profile.stored[root][0][3], profile.profile.roots[root][3], profile.profile.stored[root][0][5], profile.profile.permissions[root][3]]
    profile.profile.stored[root][1]  = crypto.hash(json.dumps([profile.profile.stored[root][0], profile.profile.stored[root][1]]))
    history = json.loads(crypto.sdec(profile.profile.consistencys[root][2], client.get(profile.profile.consistencys[root][1])))
    history.append(profile.profile.stored[root][0])
    profile.profile.consistencys[root] = mkinode(history, profile.profile.consistencys[root][0], profile.profile.consistencys[root][2])

def cd(dir):
    if dir == "..":
        dirs.pop(0)
        if dirs[0] == None:
            dirs.pop(0)
    else:
        if dir in profile.profile.roots:
            dirs.insert(0, None)
        dirs.insert(0, dir)
    if len(dirs) > 0:
        updatefilesystem()

def get(fnm):
    dirs.insert(0, fnm)
    contents = updatefilesystem()
    dirs.pop(0)
    with open(PATH_FILESYSTEM + fnm, "w") as ouf:
       ouf.write(contents[0])

def listacl():
    permissions = updatepermissions()
    for username, permission in permissions.iteritems():
        print username + ": " + permission

def ls():
    if len(dirs) == 0:
        receiveshare()
        print [dir for dir in profile.profile.roots.keys() if dir == "" or profile.profile.permissions[dir] != "owner"]
    else:
        print list(updatefilesystem()[0].keys())

def mkdir(dir):
    contents = updatefilesystem(True)
    contents[0][dir] = mkinode({})
    post(contents)

def put(fnm):
    with open(PATH_FILESYSTEM + fnm, "r") as inf:
        m = inf.read()
    contents = updatefilesystem(True)
    contents[0][fnm] = mkinode(m)
    post(contents)

def pwd():
    for i in range(len(dirs) - 1, -1, -1):
        if dirs[i] != None:
            sys.stdout.write(dirs[i] + "/")
    sys.stdout.write(os.linesep)

def rm(fnm):
    contents = updatefilesystem(True)
    client.delete(contents[0][fnm][0])
    del contents[0][fnm]
    post(contents)

def rmdir(dir):
    contents = updatefilesystem(True)
    toremove = [contents[0][dir]]
    while len(toremove) != 0:
        (efnm, hefnm, sk, hash) = toremove.pop()
        c = client.get(hefnm)
        if crypto.hash(c) != hash:
            raise Exception("ALARM: Integrity check failed on filesystem!")
        m = json.loads(crypto.sdec(sk, c))
        if isinstance(m, dict):
            toremove.extend(m.values())
        client.delete(efnm)
    del contents[0][dir]
    post(contents)

def setacl(username, permission):
    contents = None
    permissions = updatepermissions(True)
    savedpermissions = dict(permissions)
    revoke =  dirs[0] in profile.profile.roots and username in permissions and (permission == "none" or (permission == "read" and permissions[username] == "write"))
    create = (permission == "read" or permission == "write") and not dirs[0] in profile.profile.roots
    if permission == "none":
        del permissions[username]
        sendshare(username, dirs[0], permission)
    else:
        permissions[username] = permission
    if revoke or create:
        contents = updatefilesystem(True)
        if revoke and permission == "none" and len(permissions.keys()) == 2:
            client.delete(profile.profile.consistencys[dirs[0]][0])
            dirs.pop(1)
            del profile.profile.roots[dirs[0]]
            del profile.profile.permissions[dirs[0]]
            del profile.profile.consistencys[dirs[0]]
            del profile.profile.stored[dirs[0]]
        if create:
            profile.profile.roots[dirs[0]] = mkinode(contents.pop(0))
            contents[0][dirs[0]] = profile.profile.roots[dirs[0]][:]
            contents[0][dirs[0]][3] = None
            profile.profile.permissions[dirs[0]] = mkinode(permissions)
            consistency = [0, profile.profile.username, None, profile.profile.roots[dirs[0]][3], None, profile.profile.permissions[dirs[0]][3]]
            profile.profile.consistencys[dirs[0]] = mkinode([consistency])
            profile.profile.stored[dirs[0]] = [consistency, crypto.hash(json.dumps(consistency))]
            dir = dirs.pop(0)
    post(contents, savedpermissions if create else permissions)
    if create:
        dirs.insert(0, None)
        dirs.insert(0, dir)
    for username, permission in permissions.iteritems():
        if username != profile.profile.username:
            sendshare(username, dirs[0], permission)

def state():
    root = dirs[dirs.index(None) - 1]
    updateconsistencys(root, True)
    print str(profile.profile.stored[root][0][0]) + ": " + profile.profile.stored[root][1]
